from __future__ import annotations

import asyncio
import contextlib
import functools
import html
import json
import logging
import secrets
import tempfile
import typing as t
import weakref
from datetime import timedelta
from pathlib import Path
from xml.etree import ElementTree as ET

import crawlerdetect
import fastapi
import multipart
import timer_dict
from uniserde import Jsonable, JsonDoc

import rio

from .. import (
    app,
    assets,
    byte_serving,
    data_models,
    errors,
    icon_registry,
    inspection,
    routing,
    utils,
)
from ..errors import AssetError
from ..transports import FastapiWebsocketTransport, MessageRecorderTransport
from ..utils import URL
from .abstract_app_server import AbstractAppServer

try:
    import plotly  # type: ignore (missing import)
except ImportError:
    plotly = None

__all__ = [
    "FastapiServer",
]


P = t.ParamSpec("P")


# Used to identify search engine crawlers (like googlebot) and serve them
# without needing a websocket connection
CRAWLER_DETECTOR = crawlerdetect.CrawlerDetect()


@functools.lru_cache(maxsize=None)
def _build_sitemap(base_url: rio.URL, app: rio.App) -> str:
    # Build a XML site map
    tree = ET.Element(
        "urlset",
        xmlns="http://www.sitemaps.org/schemas/sitemap/0.9",
    )

    for relative_url in app._iter_page_urls(include_redirects=False):
        full_url = base_url / relative_url
        url = ET.SubElement(tree, "url")
        loc = ET.SubElement(url, "loc")
        loc.text = str(full_url)

    # Done
    return ET.tostring(tree, encoding="unicode", xml_declaration=True)


@functools.lru_cache(maxsize=None)
def read_frontend_template(template_name: str) -> str:
    """
    Read a text file from the frontend directory and return its content. The
    results are cached to avoid repeated disk access.
    """
    return (utils.FRONTEND_FILES_DIR / template_name).read_text(
        encoding="utf-8"
    )


def add_cache_headers(
    func: t.Callable[P, t.Awaitable[fastapi.Response]],
) -> t.Callable[P, t.Coroutine[None, None, fastapi.Response]]:
    """
    Decorator for routes that serve static files. Ensures that the response has
    the `Cache-Control` header set appropriately.
    """

    @functools.wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> fastapi.Response:
        response = await func(*args, **kwargs)
        response.headers["Cache-Control"] = "max-age=31536000, immutable"
        return response

    return wrapper


async def parse_uploaded_files(
    request: fastapi.Request,
) -> list[utils.FileInfo]:
    """
    Custom file upload parsing logic because FastAPI's file uploads are annoying
    in many ways:

    1. We want to return a response to the client as soon as the upload is
       complete, but then fastapi closes the files and they become unusable.
    2. Fastapi gives us *synchronous* file objects.
    3. It's unclear when the upload actually happens. Is the upload already
       complete by the time our code runs, or is the upload somehow still
       running in the background?

    Parsing the request manually gives us full control over the whole thing.
    """
    file_names = list[str]()
    file_types = list[str]()
    file_sizes = list[int]()
    files = list[tempfile.SpooledTemporaryFile]()

    _, options = multipart.parse_options_header(request.headers["content-type"])

    with multipart.PushMultipartParser(options["boundary"]) as parser:
        async for chunk in request.stream():
            for result in parser.parse(chunk):
                if isinstance(result, multipart.MultipartSegment):
                    assert result.filename
                    file_names.append(result.filename)

                    for header, value in result.headerlist:
                        if header.lower() == "content-type":
                            file_types.append(value)
                            break
                    else:
                        raise ValueError("No content-type header found")

                    file_sizes.append(0)
                    files.append(tempfile.SpooledTemporaryFile())
                elif result:  # Non-empty bytearray
                    file_sizes[-1] += len(result)
                    files[-1].write(result)
                else:  # None
                    files[-1].seek(0)

    # Build the list of file infos
    return [
        utils.FileInfo(
            name=file_names[i],
            size_in_bytes=file_sizes[i],
            media_type=file_types[i],
            contents=files[i],
        )
        for i in range(len(files))
    ]


class SpooledTempfilesTarget:
    def __init__(self):
        super().__init__()

        self.files = list[tempfile.SpooledTemporaryFile]()
        self.file_sizes = list[int]()
        self.file_names = list[str]()
        self.file_types = list[str]()

        self._current_file: tempfile.SpooledTemporaryFile | None = None
        self._current_file_size = 0

    def set_multipart_filename(self, filename: str):
        self.file_names.append(filename)

    def set_multipart_content_type(self, content_type: str):
        self.file_types.append(content_type)

    def on_data_received(self, chunk: bytes):
        if self._current_file is None:
            self._current_file = tempfile.SpooledTemporaryFile(
                mode="wb+", max_size=1_000_000
            )

        self._current_file.write(chunk)
        self._current_file_size += len(chunk)

    def on_finish(self):
        if self._current_file is None:
            return

        self._current_file.seek(0)
        self.files.append(self._current_file)
        self.file_sizes.append(self._current_file_size)

        self._current_file = None
        self._current_file_size = 0


class FastapiServer(fastapi.FastAPI, AbstractAppServer):
    def __init__(
        self,
        app_: app.App,
        debug_mode: bool,
        running_in_window: bool,
        internal_on_app_start: t.Callable[[], None] | None,
        base_url: rio.URL | None,
    ) -> None:
        super().__init__(
            title=app_.name,
            summary=None,
            description=app_.description,
            openapi_url=None,
            docs_url=None,
            redoc_url=None,
            lifespan=type(self)._lifespan,
        )
        AbstractAppServer.__init__(
            self,
            app_,
            running_in_window=running_in_window,
            debug_mode=debug_mode,
            base_url=base_url,
        )

        # If a URL was provided, run some sanity checks
        if base_url is not None:
            # If the URL is missing a protocol, yarl doesn't consider it
            # absolute. Perform a separate check for this so that the error
            # message makes more sense.
            if base_url.scheme not in ("http", "https"):
                raise ValueError(
                    "Please provide a URL that starts with either `http://` or `https://`."
                )

            # The URL must be absolute
            if not base_url.is_absolute():
                raise ValueError("The app's base URL must be absolute.")

            # The URL must not contain a query or fragment
            if base_url.query:
                raise ValueError(
                    "The app's base URL cannot contain query parameters."
                )

            if base_url.fragment:
                raise ValueError(
                    "The app's base URL cannot contain a fragment."
                )

        self.internal_on_app_start = internal_on_app_start
        self.base_url = base_url

        # While this Event is unset, no new Sessions can be created. This is
        # used to ensure that no clients (re-)connect while `rio run` is
        # reloading the app.
        self._can_create_new_sessions = asyncio.Event()
        self._can_create_new_sessions.set()

        # The session tokens and Request object for all clients that have made a
        # HTTP request, but haven't yet established a websocket connection. Once
        # the websocket connection is created, these will be turned into
        # Sessions.
        self._latent_session_tokens: dict[str, fastapi.Request] = {}

        # The session tokens for all active sessions. These allow clients to
        # identify themselves, for example to reconnect in case of a lost
        # connection.
        self._active_session_tokens: dict[str, rio.Session] = {}
        self._active_tokens_by_session: dict[rio.Session, str] = {}

        # All assets that have been registered with this session. They are held
        # weakly, meaning the session will host assets for as long as their
        # corresponding Python objects are alive.
        #
        # Assets registered here are hosted under `/asset/temp-{asset_id}`. In
        # addition the server also permanently hosts other "well known" assets
        # (such as javascript dependencies) which are available under public
        # URLS at `/asset/{some-name}`.
        self._assets: weakref.WeakValueDictionary[str, assets.Asset] = (
            weakref.WeakValueDictionary()
        )

        # All pending file uploads. These are stored in memory for a limited
        # time. When a file is uploaded the corresponding future is set.
        self._pending_file_uploads: timer_dict.TimerDict[
            str, asyncio.Future[list[utils.FileInfo]]
        ] = timer_dict.TimerDict(default_duration=timedelta(minutes=15))

        # FastAPI
        self.add_api_route("/robots.txt", self._serve_robots, methods=["GET"])
        self.add_api_route(
            "/rio/sitemap.xml", self._serve_sitemap, methods=["GET"]
        )
        self.add_api_route(
            "/rio/favicon.png", self._serve_favicon, methods=["GET"]
        )
        self.add_api_route(
            "/rio/frontend/assets/{asset_id:path}",
            self._serve_frontend_asset,
        )
        self.add_api_route(
            "/rio/assets/special/{asset_id:path}",
            self._serve_special_asset,
            methods=["GET"],
        )
        self.add_api_route(
            "/rio/assets/hosted/{asset_id:path}",
            self._serve_hosted_asset,
            methods=["GET"],
        )
        self.add_api_route(
            "/rio/assets/user/{asset_id:path}",
            self._serve_user_asset,
            methods=["GET"],
        )
        self.add_api_route(
            "/rio/assets/temp/{asset_id:path}",
            self._serve_temp_asset,
            methods=["GET"],
        )
        self.add_api_route(
            "/rio/icon/{icon_name:path}", self._serve_icon, methods=["GET"]
        )
        self.add_api_route(
            "/rio/upload/{upload_token}",
            self._serve_file_upload,
            methods=["PUT"],
        )
        self.add_api_route(
            "/rio/upload-to-component/{session_token}/{component_id}",
            self._serve_file_upload_to_component,
            methods=["PUT"],
        )
        self.add_api_websocket_route("/rio/ws", self._serve_websocket)

        # This route is only used in `debug_mode`. When the websocket connection
        # is interrupted, the frontend polls this route and then either
        # reconnects or reloads depending on whether its session token is still
        # valid (i.e. depending on whether the server was restarted or not)
        self.add_api_route(
            "/rio/validate-token/{session_token}",
            self._serve_token_validation,
        )

        # The route that serves the index.html will be registered later, so that
        # it has a lower priority than user-created routes.
        #
        # This keeps track of whether the fallback route has already been
        # registered.
        self._index_hmtl_route_registered = False

    async def __call__(self, scope, receive, send) -> None:
        # Because this is a single page application, all other routes should
        # serve the index page. The session will determine which components
        # should be shown.
        #
        # This route is registered last, so that it has the lowest priority.
        # This allows the user to add custom routes that take precedence.
        if not self._index_hmtl_route_registered:
            self._index_hmtl_route_registered = True

            self.add_api_route(
                "/{initial_route_str:path}", self._serve_index, methods=["GET"]
            )

        # Delegate to FastAPI
        return await super().__call__(scope, receive, send)

    @contextlib.asynccontextmanager
    async def _lifespan(self):
        await self._on_start()

        # Trigger any internal startup event
        if self.internal_on_app_start is not None:
            self.internal_on_app_start()

        try:
            yield
        finally:
            await self._on_close()

    def external_url_for_user_asset(self, relative_asset_path: Path) -> rio.URL:
        base_url = rio.URL("/") if self.base_url is None else self.base_url
        return base_url / f"assets/user/{relative_asset_path}"

    def weakly_host_asset(self, asset: assets.HostedAsset) -> rio.URL:
        """
        Register an asset with this server. The asset will be held weakly,
        meaning the server will host assets for as long as their corresponding
        Python objects are alive.

        If another asset with the same id is already hosted, it will be
        replaced.

        This returns the **external** URL under which the asset is accessible.
        The URL is absolute if the server had a base URL set, or otherwise
        relative (i.e. starting with `/`).
        """
        self._assets[asset.secret_id] = asset
        base_url = rio.URL("/") if self.base_url is None else self.base_url
        return base_url / f"rio/assets/temp/{asset.secret_id}"

    def _get_all_meta_tags(self, title: str | None = None) -> list[str]:
        """
        Returns all `<meta>` tags that should be added to the app's HTML page.
        This includes auto-generated ones, as well as those stored directly in
        the app.
        """
        if title is None:
            title = self.app.name

        base_url = rio.URL("/") if self.base_url is None else self.base_url

        # Prepare the default tags
        all_tags = {
            "og:title": title,
            "description": self.app.description,
            "og:description": self.app.description,
            "keywords": "python, web, app, rio",
            "image": str(base_url / "rio/favicon.png"),
            "viewport": "width=device-width, height=device-height, initial-scale=1, minimum-scale=1",
        }

        # Add the user-defined meta tags, overwriting any automatically
        # generated ones
        all_tags.update(self.app._custom_meta_tags)

        # Convert everything to HTML
        result: list[str] = []

        for key, value in all_tags.items():
            key = html.escape(key)
            value = html.escape(value)
            result.append(f'<meta name="{key}" content="{value}">')

        return result

    async def _serve_index(
        self,
        request: fastapi.Request,
        initial_route_str: str,
    ) -> fastapi.responses.Response:
        """
        Handler for serving the index HTML page via fastapi.
        """
        # Because Rio apps are single-page, this route serves as the fallback.
        # In addition to legitimate requests for HTML pages, it will also catch
        # a bunch of invalid requests to other resources. To highlight this,
        # throw a 404 if HTML is not explicitly requested.
        #
        # Currently inactive, because this caused issues behind dumb proxies
        # that don't pass on the `accept` header.

        # if not request.headers.get("accept", "").startswith("text/html"):
        #     raise fastapi.HTTPException(
        #         status_code=fastapi.status.HTTP_404_NOT_FOUND,
        #     )

        initial_messages = list[JsonDoc]()

        is_crawler = CRAWLER_DETECTOR.isCrawler(
            request.headers.get("User-Agent")
        )
        if is_crawler:
            # If it's a crawler, immediately create a Session for it. Instead of
            # a websocket connection, outgoing messages are simply appended to a
            # list which will be included in the HTML.
            transport = MessageRecorderTransport()
            initial_messages = transport.sent_messages

            assert request.client is not None, "How can this happen?!"

            requested_url = rio.URL(str(request.url))
            try:
                session = await self.create_session(
                    initial_message=data_models.InitialClientMessage.from_defaults(
                        url=str(requested_url),
                    ),
                    transport=transport,
                    client_ip=request.client.host,
                    client_port=request.client.port,
                    http_headers=request.headers,
                )
            except routing.NavigationFailed:
                raise fastapi.HTTPException(
                    status_code=fastapi.status.HTTP_500_INTERNAL_SERVER_ERROR,
                    detail=f"Navigation to initial page `{request.url}` has failed.",
                ) from None

            session.close()

            # If a page guard caused a redirect, tell that to the crawler in a
            # language it understands
            if session.active_page_url != requested_url:
                return fastapi.responses.RedirectResponse(
                    str(session.active_page_url)
                )

            session_token = "<crawler>"

            title = " - ".join(
                page.name for page in session.active_page_instances
            )
        else:
            # Create a session token that uniquely identifies this client
            session_token = secrets.token_urlsafe()
            self._latent_session_tokens[session_token] = request

            title = self.app.name

        # Load the template
        html_ = read_frontend_template("index.html")

        html_ = html_.replace(
            "{session_token}",
            session_token,
        )

        html_ = html_.replace(
            '"{child_attribute_names}"',
            json.dumps(
                inspection.get_child_component_containing_attribute_names_for_builtin_components()
            ),
        )

        html_ = html_.replace(
            '"{ping_pong_interval}"',
            str(self.app._ping_pong_interval.total_seconds()),
        )

        html_ = html_.replace(
            '"{debug_mode}"',
            "true" if self.debug_mode else "false",
        )

        html_ = html_.replace(
            '"{running_in_window}"',
            "true" if self.running_in_window else "false",
        )

        html_ = html_.replace(
            '"{initial_messages}"',
            json.dumps(initial_messages),
        )

        if self.base_url is None:
            html_base_url = "/"
        else:
            html_base_url = str(self.base_url)
            html_base_url = html_base_url.rstrip("/") + "/"

        html_ = html_.replace(
            "/rio-base-url-placeholder/",
            html_base_url,
        )

        theme = self.app._theme
        if isinstance(theme, tuple):
            light_theme_background_color = theme[0].background_color
            dark_theme_background_color = theme[1].background_color
        else:
            light_theme_background_color = theme.background_color
            dark_theme_background_color = theme.background_color

        html_ = html_.replace(
            "{light_theme_background_color}",
            f"#{light_theme_background_color.hexa}",
        )

        html_ = html_.replace(
            "{dark_theme_background_color}",
            f"#{dark_theme_background_color.hexa}",
        )

        # Since the title is user-defined, it might contain placeholders like
        # `{debug_mode}`. So it's important that user-defined content is
        # inserted last.
        html_ = html_.replace(
            "{title}",
            html.escape(title),
        )

        # The placeholder for the metadata uses unescaped `<` and `>` characters
        # to ensure that no user-defined content can accidentally contain this
        # placeholder.
        html_ = html_.replace(
            '<meta name="{meta}" />',
            "\n".join(self._get_all_meta_tags(title)),
        )

        # Respond
        return fastapi.responses.HTMLResponse(html_)

    async def _serve_robots(
        self, request: fastapi.Request
    ) -> fastapi.responses.Response:
        """
        Handler for serving the `robots.txt` file via fastapi.
        """
        # Under which URL is the app hosted?
        if self.base_url is None:
            base_url = URL(str(request.url))
        else:
            base_url = self.base_url

        # FIXME: Disallow internal API routes? Icons, assets, etc?
        content = f"""
User-agent: *
Allow: /

Sitemap: {base_url / "rio/sitemap.xml"}
        """.strip()

        return fastapi.responses.Response(
            content=content,
            media_type="text/plain",
        )

    async def _serve_sitemap(
        self, request: fastapi.Request
    ) -> fastapi.responses.Response:
        """
        Handler for serving the `sitemap.xml` file via fastapi.
        """
        # Under which URL is the app hosted?
        if self.base_url is None:
            base_url = URL(str(request.url)).origin()
        else:
            base_url = self.base_url

        # Respond
        return fastapi.responses.Response(
            content=_build_sitemap(base_url, self.app),
            media_type="application/xml",
        )

    async def _serve_favicon(self) -> fastapi.responses.Response:
        """
        Handler for serving the favicon via fastapi, if one is set.
        """
        # Fetch the favicon. This method is already caching, so it's fine to
        # fetch every time.
        try:
            icon_png_blob = await self.app._fetch_icon_png_blob()
        except IOError as err:
            logging.error(str(err))

            raise fastapi.HTTPException(
                status_code=fastapi.status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail="Could not load the app icon.",
            ) from err

        # Respond
        return fastapi.responses.Response(
            content=icon_png_blob,
            media_type="image/png",
        )

    async def _serve_frontend_asset(
        self,
        request: fastapi.Request,
        asset_id: str,
    ) -> fastapi.responses.Response:
        response = await self._serve_file_from_directory(
            request,
            utils.FRONTEND_ASSETS_DIR,
            asset_id + ".gz",
        )
        response.headers["content-encoding"] = "gzip"

        return response

    @add_cache_headers
    async def _serve_special_asset(
        self,
        request: fastapi.Request,
        asset_id: str,
    ) -> fastapi.responses.Response:
        # The python plotly library already includes a minified version of
        # plotly.js. Rather than shipping another one, just serve the one
        # included in the library.
        if asset_id == "plotly.min.js":
            if plotly is None:
                return fastapi.responses.Response(status_code=404)

            return fastapi.responses.Response(
                content=plotly.offline.get_plotlyjs(),
                media_type="text/javascript",
            )

        return fastapi.responses.Response(status_code=404)

    async def _serve_hosted_asset(
        self,
        request: fastapi.Request,
        asset_id: str,
    ) -> fastapi.responses.Response:
        return await self._serve_file_from_directory(
            request,
            utils.HOSTED_ASSETS_DIR,
            asset_id,
        )

    async def _serve_user_asset(
        self,
        request: fastapi.Request,
        asset_id: str,
    ) -> fastapi.responses.Response:
        return await self._serve_file_from_directory(
            request,
            self.app.assets_dir,
            asset_id,
        )

    @add_cache_headers
    async def _serve_temp_asset(
        self,
        request: fastapi.Request,
        asset_id: str,
    ) -> fastapi.responses.Response:
        # Get the asset's Python instance. The asset's id acts as a secret, so
        # no further authentication is required.
        try:
            asset = self._assets[asset_id]
        except KeyError:
            return fastapi.responses.Response(status_code=404)

        # Fetch the asset's content and respond
        if isinstance(asset, assets.BytesAsset):
            return fastapi.responses.Response(
                content=asset.data,
                media_type=asset.media_type,
            )
        elif isinstance(asset, assets.PathAsset):
            return byte_serving.range_requests_response(
                request,
                file_path=asset.path,
                media_type=asset.media_type,
            )
        else:
            assert False, f"Unable to serve asset of unknown type: {asset}"

    @add_cache_headers
    async def _serve_file_from_directory(
        self,
        request: fastapi.Request,
        directory: Path,
        asset_id: str,
    ) -> fastapi.responses.Response:
        # Construct the path to the target file
        asset_file_path = directory / asset_id

        # Make sure the path is inside the hosted assets directory
        asset_file_path = asset_file_path.absolute()
        if not asset_file_path.is_relative_to(directory):
            logging.warning(
                f'Client requested asset "{asset_id}" which is not located'
                f" inside the assets directory. Somebody might be trying to"
                f" break out of the assets directory!"
            )
            return fastapi.responses.Response(status_code=404)

        return byte_serving.range_requests_response(
            request,
            file_path=asset_file_path,
        )

    @add_cache_headers
    async def _serve_icon(self, icon_name: str) -> fastapi.responses.Response:
        """
        Allows the client to request an icon by name. This is not actually the
        mechanism used by the `Icon` component, but allows JavaScript to request
        icons.
        """
        # Get the icon's SVG
        try:
            svg_source = icon_registry.get_icon_svg(icon_name)
        except AssetError:
            return fastapi.responses.Response(status_code=404)

        # Respond
        return fastapi.responses.Response(
            content=svg_source,
            media_type="image/svg+xml",
        )

    async def _serve_file_upload(
        self, upload_token: str, request: fastapi.Request
    ) -> fastapi.Response:
        # Try to find the future for this token
        try:
            future = self._pending_file_uploads.pop(upload_token)
        except KeyError:
            raise fastapi.HTTPException(
                status_code=fastapi.status.HTTP_400_BAD_REQUEST,
                detail="Invalid upload token.",
            )

        # Complete the future
        files = await parse_uploaded_files(request)
        future.set_result(files)

        return fastapi.responses.Response(
            status_code=fastapi.status.HTTP_200_OK
        )

    async def _serve_file_upload_to_component(
        self,
        session_token: str,
        component_id: str,
        request: fastapi.Request,
    ) -> fastapi.Response:
        # Parse the inputs
        try:
            component_id_int = int(component_id)
        except ValueError:
            raise fastapi.HTTPException(
                status_code=fastapi.status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail="Invalid component ID.",
            )

        # Find the session for this token
        try:
            session = self._active_session_tokens[session_token]
        except KeyError:
            raise fastapi.HTTPException(
                status_code=fastapi.status.HTTP_401_UNAUTHORIZED,
                detail="Invalid session token.",
            )

        # Find the component in this session
        #
        # This can reasonably happen without it being an error - the component
        # could've been deleted in the meantime.
        try:
            component = session._weak_components_by_id[component_id_int]
        except KeyError:
            return fastapi.responses.Response(
                status_code=fastapi.status.HTTP_200_OK
            )

        # Does this component even have a handler for file uploads?
        try:
            handler = component._on_file_upload_  # type: ignore
        except AttributeError:
            raise fastapi.HTTPException(
                status_code=fastapi.status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail="This component does not accept file uploads.",
            )

        # Get all uploaded files
        files = await parse_uploaded_files(request)

        # Let the component handle the files
        await handler(files)

        return fastapi.responses.Response(
            status_code=fastapi.status.HTTP_200_OK
        )

    async def pick_file(
        self,
        session: rio.Session,
        *,
        file_types: list[str] | None = None,
        multiple: bool = False,
    ) -> utils.FileInfo | list[utils.FileInfo]:
        # Create a secret id and register the file upload with the app server
        upload_id = secrets.token_urlsafe()
        future = asyncio.Future[list[utils.FileInfo]]()

        self._pending_file_uploads[upload_id] = future

        # Tell the frontend to upload a file
        base_url = rio.URL("/") if self.base_url is None else self.base_url

        await session._request_file_upload(
            upload_url=str(base_url / f"rio/upload/{upload_id}"),
            file_types=file_types,
            multiple=multiple,
        )

        # Wait for the user to upload files
        files = await future

        # Raise an exception if no files were uploaded
        if not files:
            raise errors.NoFileSelectedError()

        # Ensure only one file was provided if `multiple` is False
        if not multiple and len(files) != 1:
            logging.warning(
                "Client attempted to upload multiple files when `multiple` was False."
            )
            raise errors.NoFileSelectedError()

        # Return the file info
        if multiple:
            return files
        else:
            return files[0]

    async def _serve_token_validation(
        self, request: fastapi.Request, session_token: str
    ) -> fastapi.Response:
        return fastapi.responses.JSONResponse(
            session_token in self._active_session_tokens
        )

    @contextlib.contextmanager
    def temporarily_disable_new_session_creation(self):
        self._can_create_new_sessions.clear()

        try:
            yield
        finally:
            self._can_create_new_sessions.set()

    async def _serve_websocket(
        self,
        websocket: fastapi.WebSocket,
        sessionToken: str,
    ) -> None:
        """
        Handler for establishing the websocket connection and handling any
        messages.
        """
        # Blah, naming conventions
        session_token = sessionToken
        del sessionToken

        rio._logger.debug(
            f"Received websocket connection with session token `{session_token}`"
        )

        # Wait until we're allowed to accept new websocket connections. (This is
        # temporarily disable while `rio run` is reloading the app.)
        await self._can_create_new_sessions.wait()

        # Accept the socket. I can't figure out how to access the close reason
        # in JS unless the connection was accepted first, so we have to do this
        # even if we are going to immediately close it again.
        await websocket.accept()

        # Look up the session token. If it is valid the session's duration is
        # refreshed so it doesn't expire. If the token is not valid, don't
        # accept the websocket.
        try:
            request = self._latent_session_tokens.pop(session_token)
        except KeyError:
            # Check if this is a reconnect
            try:
                sess = self._active_session_tokens[session_token]
            except KeyError:
                # Inform the client that this session token is invalid
                await websocket.close(
                    3000,  # Custom error code
                    "Invalid session token.",
                )
                return

            # Check if this session still has a functioning websocket
            # connection. Browsers have a "duplicate tab" feature that can
            # create a 2nd tab with the same session token as the original one,
            # and in that case we want to create a new session.
            if sess._transport is not None:
                await websocket.close(
                    3000,  # Custom error code
                    "Invalid session token.",
                )
                return

            # Replace the session's websocket
            sess._transport = transport = FastapiWebsocketTransport(websocket)

            # Make sure the client is in sync with the server by refreshing
            # every single component
            await sess._send_all_components_on_reconnect()

            # Start listening for incoming messages. The session has just
            # received a new websocket connection, so a new task is needed.
            #
            # The previous one was closed when the transport was replaced.
            self._session_serve_tasks[sess] = asyncio.create_task(
                self._serve_session(sess),
                name=f"`Session.serve` for session id `{id(sess)}`",
            )

        else:
            transport = FastapiWebsocketTransport(websocket)

            try:
                sess = await self._create_session_from_websocket(
                    session_token, request, websocket, transport
                )
            except fastapi.WebSocketDisconnect:
                # If the websocket disconnected while we were initializing the
                # session, just close it
                return

            self._active_session_tokens[session_token] = sess
            self._active_tokens_by_session[sess] = session_token

            # Trigger a refresh. This will also send the initial state to
            # the frontend.
            await sess._refresh()

        # Apparently the websocket becomes unusable as soon as this function
        # exits, so we must wait until we no longer need the websocket.
        #
        # When exiting `rio run` with Ctrl+C, this task is cancelled and screams
        # loudly in the console. Suppress that by catching the exception.
        try:
            await transport.closed.wait()
        except asyncio.CancelledError:
            pass

    async def _create_session_from_websocket(
        self,
        session_token: str,
        request: fastapi.Request,
        websocket: fastapi.WebSocket,
        transport: FastapiWebsocketTransport,
    ) -> rio.Session:
        assert request.client is not None, "Why can this happen?"

        # Upon connecting, the client sends an initial message containing
        # information about it. Wait for that, but with a timeout - otherwise
        # evildoers could overload the server with connections that never send
        # anything.
        initial_message_json: Jsonable = await asyncio.wait_for(
            websocket.receive_json(),
            timeout=60,
        )

        initial_message = data_models.InitialClientMessage.from_json(
            initial_message_json  # type: ignore
        )

        try:
            sess = await self.create_session(
                initial_message,
                transport=transport,
                client_ip=request.client.host,
                client_port=request.client.port,
                http_headers=request.headers,
            )
        except routing.NavigationFailed:
            # TODO: Notify the client? Show an error?
            raise fastapi.HTTPException(
                status_code=fastapi.status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail=f"Navigation to initial page `{request.url}` has failed.",
            ) from None

        return sess

    def _after_session_closed(self, session: rio.Session) -> None:
        super()._after_session_closed(session)

        try:
            session_token = self._active_tokens_by_session.pop(session)
        except KeyError:
            # It must be a session created for a crawler. Those don't get unique
            # session tokens and aren't registered.
            return

        del self._active_session_tokens[session_token]



from __future__ import annotations

import pytz
import starlette.datastructures

import rio

from .. import assets, utils
from ..transports import MessageRecorderTransport
from .abstract_app_server import AbstractAppServer

__all__ = ["TestingServer"]


class TestingServer(AbstractAppServer):
    def __init__(
        self,
        app: rio.App,
        *,
        running_in_window: bool = False,
        debug_mode: bool = False,
        base_url: rio.URL | None = None,
    ) -> None:
        super().__init__(
            app,
            running_in_window=running_in_window,
            debug_mode=debug_mode,
            base_url=base_url,
        )

    def create_dummy_session(self) -> rio.Session:
        url = rio.URL("http://localhost")

        theme = self.app._theme
        if isinstance(theme, tuple):
            theme = theme[0]

        return rio.Session(
            app_server_=self,
            transport=MessageRecorderTransport(),
            client_ip="localhost",
            client_port=1234,
            http_headers=starlette.datastructures.Headers(),
            timezone=pytz.UTC,
            preferred_languages=["en-US"],
            month_names_long=(
                "January",
                "February",
                "March",
                "April",
                "May",
                "June",
                "July",
                "August",
                "September",
                "October",
                "November",
                "December",
            ),
            day_names_long=(
                "Monday",
                "Tuesday",
                "Wednesday",
                "Thursday",
                "Friday",
                "Saturday",
                "Sunday",
            ),
            date_format_string="",
            first_day_of_week=0,
            decimal_separator=".",
            thousands_separator=",",
            screen_width=800,
            screen_height=600,
            window_width=800,
            window_height=600,
            pixels_per_font_height=16,
            scroll_bar_size=16,
            primary_pointer_type="mouse",
            base_url=url,
            active_page_url=url,
            theme_=theme,
        )

    def weakly_host_asset(self, asset: assets.HostedAsset) -> rio.URL:
        # This method is called for font files, even in a test session, so it
        # actually needs to be implemented.
        return rio.URL(asset.secret_id)

    def external_url_for_user_asset(
        self, relative_asset_path: assets.Path
    ) -> rio.URL:
        raise NotImplementedError

    async def pick_file(
        self,
        session: rio.Session,
        *,
        file_types: list[str] | None = None,
        multiple: bool = False,
    ) -> utils.FileInfo | list[utils.FileInfo]:
        raise NotImplementedError



# <additional-imports>
import functools
import typing as t

import rio

from .. import data_models

# </additional-imports>


# <component>
@rio.page(
    name="CRUD",
    url_segment="",
)
class CrudPage(rio.Component):
    """
    A CRUD page that allows users to create, read, update, and delete menu
    items.

    The @rio.event.on_populate decorator is used to fetch data from a predefined
    data model and assign it to the menu_items attribute of the current
    instance.


    ## Attributes

    `menu_items`: A list of menu items.

    `currently_selected_menu_item`: The currently selected menu item.

    `banner_text`: The text to be displayed in the banner.

    `banner_style`: The style of the banner (success, danger, info).
    """

    menu_items: list[data_models.MenuItem] = []
    currently_selected_menu_item: data_models.MenuItem | None = None
    banner_text: str = ""
    banner_style: t.Literal["success", "danger", "info"] = "success"

    @rio.event.on_populate
    def on_populate(self) -> None:
        """
        Event handler that is called when the component is populated.

        Fetches data from a predefined data model and assigns it to the menu_items
        attribute of the current instance.
        """
        self.menu_items = data_models.MENU_ITEMS

    async def on_press_delete_item(self, idx: int) -> None:
        """
        Perform actions when the "Delete" button is pressed.

        ## Parameters

        `idx`: The index of the item to be deleted.
        """
        # delete the item from the list
        self.menu_items.pop(idx)
        self.banner_text = "Item was deleted"
        self.banner_style = "danger"
        self.currently_selected_menu_item = None

    # Helper function to create a dialog for editing a menu item
    async def _create_dialog_item_editor(
        self, selected_menu_item: data_models.MenuItem, new_entry: bool
    ) -> data_models.MenuItem | None:
        """
        Creates a dialog to edit or add a menu item.

        This method creates a dialog that allows the user to edit or add a menu
        item. The dialog contains input fields for the name, description, price,
        and category of the menu item. The user can save or cancel the changes.
        If the user saves the changes, the updated menu item is returned. If the
        user cancels the changes, the original menu item is returned.

        ## Parameters

        `selected_menu_item`: The selected menu item to be edited or added.

        `new_entry`: A boolean flag indicating if the menu item is a new entry.


        See the approx. layout below:

        ```
        ╔══════════════════════ Card ══════════════════════╗
        ║ ┏━━━━━━━━━━━━━━━━━━━━ Text ━━━━━━━━━━━━━━━━━━━━┓ ║
        ║ ┃ Edit Menu Item | Add New Menu Item           ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ║ ┏━━━━━━━━━━━━━━━━━━ TextInput ━━━━━━━━━━━━━━━━━┓ ║
        ║ ┃ Name                                         ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ║ ┏━━━━━━━━━━━━━━━━━━ TextInput ━━━━━━━━━━━━━━━━━┓ ║
        ║ ┃ Description                                  ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ║ ┏━━━━━━━━━━━━━━━━━ NumberInput ━━━━━━━━━━━━━━━━┓ ║
        ║ ┃ Price                                        ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ║ ┏━━━━━━━━━━━━━━━━━━ Dropdown ━━━━━━━━━━━━━━━━━━┓ ║
        ║ ┃ Category                                     ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ║ ┏━━━━━━━━━━━━━━━━━━━━━ Row ━━━━━━━━━━━━━━━━━━━━┓ ║
        ║ ┃   ┏━━━ Button ━━━┓        ┏━━━ Button ━━━┓   ┃ ║
        ║ ┃   ┃ Save         ┃        ┃ Cancel       ┃   ┃ ║
        ║ ┃   ┗━━━━━━━━━━━━━━┛        ┗━━━━━━━━━━━━━━┛   ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ╚══════════════════════════════════════════════════╝
        ```
        """
        # Make a copy of the selected menu item to avoid modifying the original,
        # which is returned if the user cancels the dialog.
        selected_menu_item_copied = selected_menu_item.copy()

        # This function will be called to create the dialog's content.
        # It builds up a UI using Rio components, just like a regular
        # `build` function would.
        def build_dialog_content() -> rio.Component:
            # Build the dialog
            if new_entry is False:
                text = "Edit Menu Item"
            else:
                text = "Add New Menu Item"

            return rio.Column(
                rio.Text(
                    text=text,
                    style="heading2",
                    margin_bottom=1,
                ),
                rio.TextInput(
                    selected_menu_item_copied.name,
                    label="Name",
                    on_change=on_change_name,
                ),
                rio.TextInput(
                    selected_menu_item_copied.description,
                    label="Description",
                    on_change=on_change_description,
                ),
                rio.NumberInput(
                    selected_menu_item_copied.price,
                    label="Price",
                    suffix_text="$",
                    on_change=on_change_price,
                ),
                rio.Dropdown(
                    options=[
                        "Burgers",
                        "Desserts",
                        "Drinks",
                        "Salads",
                        "Sides",
                    ],
                    label="Category",
                    selected_value=selected_menu_item_copied.category,
                    on_change=on_change_category,
                ),
                rio.Row(
                    rio.Button(
                        "Save",
                        on_press=lambda selected_menu_item_copied=selected_menu_item_copied: dialog.close(
                            selected_menu_item_copied
                        ),
                    ),
                    rio.Button(
                        "Cancel",
                        on_press=lambda: dialog.close(selected_menu_item),
                    ),
                    spacing=1,
                    align_x=1,
                ),
                spacing=1,
                align_y=0,
                margin=2,
                min_width=30,
            )

        def on_change_name(ev: rio.TextInputChangeEvent) -> None:
            """
            Changes the name of the currently selected menu item. And updates the
            name attribute of our data model.

            ## Parameters

            `ev`: The event object that contains the new name.
            """
            selected_menu_item_copied.name = ev.text

        def on_change_description(
            ev: rio.TextInputChangeEvent,
        ) -> None:
            """
            Changes the description of the currently selected menu item. And updates
            the description attribute of our data model.

            ## Parameters

            `ev`: The event object that contains the new description.
            """
            selected_menu_item_copied.description = ev.text

        def on_change_price(ev: rio.NumberInputChangeEvent) -> None:
            """
            Changes the price of the currently selected menu item. And updates the
            price attribute of our data model.

            ## Parameters

            `ev`: The event object that contains the new price.
            """
            selected_menu_item_copied.price = ev.value

        def on_change_category(ev: rio.DropdownChangeEvent) -> None:
            """
            Changes the category of the currently selected menu item. And updates
            the category attribute of our data model.

            ## Parameters

            `ev`: The event object that contains the new category.
            """
            selected_menu_item_copied.category = ev.value

        # Show the dialog
        dialog = await self.session.show_custom_dialog(
            build=build_dialog_content,
            # Prevent the user from interacting with the rest of the app
            # while the dialog is open
            modal=True,
            # Don't close the dialog if the user clicks outside of it
            user_closable=False,
        )

        # Wait for the user to select an option
        result = await dialog.wait_for_close()

        # Return the selected value
        return result

    async def on_spawn_dialog_edit_menu_item(
        self, selected_menu_item: data_models.MenuItem, idx: int
    ) -> None:
        """
        Opens a dialog to edit the selected menu item.

        Updates the menu item at the given index if the user confirms the changes.

        ## Parameters

        `selected_menu_item`: The selected menu item to be edited.

        `idx`: The index of the selected menu item in the list of menu items.
        """
        assert selected_menu_item is not None
        result = await self._create_dialog_item_editor(
            selected_menu_item=selected_menu_item, new_entry=False
        )

        # Ensure the result is not None
        if result is None:
            self.banner_text = "Item was **NOT** updated"
            self.banner_style = "danger"
        else:
            # Update the menu item
            self.menu_items[idx] = result
            self.banner_text = "Item was updated"
            self.banner_style = "info"

    async def on_spawn_dialog_add_new_menu_item(self) -> None:
        """
        Perform actions when the "Add New" ListItem is pressed.

        This method creates a new empty menu item of models.MenuItems.
        It then opens a dialog for the user to enter the details of the
        new menu item. If the user confirms the addition and the new
        menu item is not empty, it appends the new menu item to the list
        of menu items and updates the banner text accordingly.

        If the user cancels the addition or the new menu item is empty,
        it updates the banner text to indicate that the item was not added.
        """
        new_menu_item = data_models.MenuItem.new_empty()
        result = await self._create_dialog_item_editor(
            selected_menu_item=new_menu_item, new_entry=True
        )

        # Ensure the result is not None
        if result is None:
            self.banner_text = "Item was NOT updated"
            self.banner_style = "danger"
        else:
            # Append the new menu item to our list of menu items only
            # if it is not empty
            if result != data_models.MenuItem.new_empty():
                self.menu_items.append(result)
                self.banner_text = "Item was added"
                self.banner_style = "success"
            else:
                self.banner_text = "Item was NOT added"
                self.banner_style = "danger"

    def build(self) -> rio.Component:
        """
        Builds the component to be rendered.

        If there is no currently selected menu item, only the Banner and
        ItemList component is returned.

        When you click on a SimpleListItem, a custom Dialog appears, allowing
        you to edit the selected item. Similarly, clicking on the "Add new"
        SimpleListItem opens a custom Dialog for adding a new item.

        See the approx. layout below:

        ```
        ╔══════════════════════ Card ═════════════════════════╗
        ║ ┏━━━━━━━━━━━━━━━━━━━━ Banner ━━━━━━━━━━━━━━━━━━━━━┓ ║
        ║ ┃ "" | Item was updated | Item was added          ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ║ ┏━━━━━━━━━━━━━━━━━━━ ListView ━━━━━━━━━━━━━━━━━━━━┓ ║
        ║ ┃ ┏━━━━━━━━━━━━━━━━━ SimpleListItem ━━━━━━━━━━━━┓ ┃ ║
        ║ ┃ ┃ Add new                                     ┃ ┃ ║
        ║ ┃ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ┃ ║
        ║ ┃ ┏━━━━━━━━━━━━━━━━━ SimpleListItem ━━━━━━━━━━━━┓ ┃ ║
        ║ ┃ ┃ Item 1                          ┏━Button━┓  ┃ ┃ ║
        ║ ┃ ┃                                 ┗━━━━━━━━┛  ┃ ┃ ║
        ║ ┃ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ┃ ║
        ║ ┃ ┏━━━━━━━━━━━━━━━━━ SimpleListItem ━━━━━━━━━━━━┓ ┃ ║
        ║ ┃ ┃ Item 2                          ┏━Button━┓  ┃ ┃ ║
        ║ ┃ ┃                                 ┗━━━━━━━━┛  ┃ ┃ ║
        ║ ┃ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ┃ ║
        ║ ┃ ...                                             ┃ ║
        ║ ┃ ┏━━━━━━━━━━━━━━━━━ SimpleListItem ━━━━━━━━━━━━┓ ┃ ║
        ║ ┃ ┃ Item n                          ┏━Button━┓  ┃ ┃ ║
        ║ ┃ ┃                                 ┗━━━━━━━━┛  ┃ ┃ ║
        ║ ┃ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ╚═════════════════════════════════════════════════════╝
        ```
        """

        # Store all children in an intermediate list
        list_items = []

        list_items.append(
            rio.SimpleListItem(
                text="Add new",
                secondary_text="Description",
                key="add_new",
                left_child=rio.Icon("material/add"),
                on_press=self.on_spawn_dialog_add_new_menu_item,
            )
        )

        for i, item in enumerate(self.menu_items):
            list_items.append(
                rio.SimpleListItem(
                    text=item.name,
                    secondary_text=item.description,
                    right_child=rio.Button(
                        rio.Icon("material/delete", margin=0.5),
                        color=self.session.theme.danger_color,
                        min_width=8,
                        # Note the use of functools.partial to pass the
                        # index to the event handler.
                        on_press=functools.partial(
                            self.on_press_delete_item, i
                        ),
                    ),
                    # Use the name as the key to ensure that the list item
                    # is unique.
                    key=item.name,
                    # Note the use of functools.partial to pass the
                    # item to the event handler.
                    on_press=functools.partial(
                        self.on_spawn_dialog_edit_menu_item, item, i
                    ),
                )
            )

        # Then unpack the list to pass the children to the ListView
        return rio.Column(
            rio.Banner(
                self.banner_text,
                style=self.banner_style,
                margin_bottom=1,
            ),
            rio.ListView(
                *list_items,
                align_y=0,
            ),
            # align at the top
            align_y=0,
            margin=3,
        )


# </component>



from __future__ import annotations

import dataclasses
import typing as t

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "Switcher",
]


@t.final
class Switcher(FundamentalComponent):
    """
    Smoothly transitions between components.

    The `Switcher` component is a container which can display one component at a
    time. What makes it useful, is that when you change the `content` attribute,
    rather than instantly swapping the displayed component, it will smoothly
    transition between the two.

    Moreover, whenever the content's size changes, the `Switcher` will
    smoothly resize to match the new size. This means you can use switchers
    to smoothly transition between components of different sizes.

    `content` may also be `None`, in which case the `Switcher` won't display
    anything. This in turn allows you to animate the appearance or disappearance
    of a component, e.g. for a sidebar.


    ## Attributes

    `content`: The component to display inside the switcher. If `None`, the
        switcher will be empty.

    `transition_time`: How many seconds it should take for the switcher to
        transition between components and sizes.


    ## Example

    This example showcases a simple switcher component that allows users to
    control the visibility of a content element. The content element is a
    rectangle that appears or disappears when the button is pressed.

    ```python
    class MyComponent(rio.Component):
        show_content: bool = False

        def _on_press(self) -> None:
            self.show_content = not self.show_content

        def build(self) -> rio.Component:
            # Make the content appear or disappear based on the show_content
            # attribute
            content = (
                rio.Rectangle(
                    fill=rio.Color.from_hex("00bf63"),
                    min_height=12,
                )
                if self.show_content
                else None  # No content if show_content is False
            )
            return rio.Column(
                rio.Button("Show content", on_press=self._on_press),
                # Switcher to show/hide content
                rio.Switcher(
                    content=content,
                ),
                align_y=0.5,
                align_x=0.5,
            )
    ```
    ## Metadata

    `experimental`: True
    """

    content: rio.Component | None

    _: dataclasses.KW_ONLY

    transition_time: float = 0.35


Switcher._unique_id_ = "Switcher-builtin"



from __future__ import annotations

from pathlib import Path

import rio

# This boolean indicates whether the program was started via `rio run`. This
# lets us display more useful error messages if the user tries to execute
# something like `app.run_in_window()`, which conflicts with `rio run`.
launched_via_rio_run: bool = False


# When launched via `rio run`, the usual ways to detect the "main file" (`import
# __main__` and `sys.argv`) don't work. So `rio run` explicitly tells us what
# the "main file" is by setting this variable.
rio_run_app_module_path: Path | None = None


# Before a component is built, this value is set to that component. This allows
# newly created components to determine their creator, as well as session.
#
# - `Component`: The component that is currently being built
# - `None`: The build method of the app itself is currently being called
currently_building_component: rio.Component | None = None


# Same as `currently_building_component`, but holding that component's session.
#
# - `Session`: The session that owns the currently building component
# - `None`: No build is currently in progress
currently_building_session: rio.Session | None = None


# This counter is increased each time a component is built. It can thus be used to
# uniquely identify the build generation of every component.
build_generation: int = 0
